(*
 * Copyright (c) 2002 Andr di Geronimo (adg@home.se)
 *)
 
unit O_geo;

interface

uses
  Windows, Classes, SysUtils,
  P_vec;

type
  TBaseList = class;
  T3DMesh = class;
  T3DMeshList = class;
  TSurface = class;
  TSurfaceList = class;
  TVertex = class;
  TVertexList = class;
  TTXVertex = class;
  TTXVertexList = class;

  TBaseList = class(TList)
  public
    destructor Destroy; override;
    procedure FreeAll;
  end;

  T3DMesh = class(TObject)
  private
    FVertices : TVertexList;
    FSurfaces : TSurfaceList;
  public
    constructor Create;
    destructor Destroy; override;
    property Surfaces : TSurfaceList read FSurfaces;
    property Vertices : TVertexList read FVertices;
  end;

  T3DMeshList = class(TBaseList)
  protected
    function GetItem(i : integer) : T3DMesh;
    procedure SetItem(i : integer; msh : T3DMesh);
  public
    property Items[i : integer] : T3DMesh read GetItem write SetItem; default;
  end;

  TSurface = class(TObject)
  private
    FVertices : TVertexList;
    FTXVertices : TTXVertexList;
    FvecNormal : TVector;
    FiAdjoin : integer;
    FiPortalSec : integer;
    FiPortalSur : integer;
    FsMaterial : string;
    FiFlags : integer;
//    FvecCenter: TVector;
//    FbCenterCalculated : boolean;
    function GetIsPortal: boolean;
    function GetCenter: TVector;
    function GetNormal: TVector;
  public
    constructor Create;
    destructor Destroy; override;
    function AddVertex(vx : TVertex ; txvx : TTXVertex = nil) : integer;
    procedure DeleteVertex(i : integer);
    property Vertices : TVertexList read FVertices;
    property TXVertices : TTXVertexList read FTXVertices;
    property iAdjoin : integer read FiAdjoin write FiAdjoin;
    property iPortalSec : integer read FiPortalSec write FiPortalSec;
    property iPortalSur : integer read FiPortalSur write FiPortalSur;
    property bIsPortal : boolean read GetIsPortal;
    property sMaterial : string read FsMaterial write FsMaterial;
    property Flags : integer read FiFlags write FiFlags;
    property vecNormal : TVector read GetNormal write FvecNormal;
    property vecCenter : TVector read GetCenter;
  end;

  TSurfaceList = class(TBaseList)
  protected
    function GetItem(i : integer) : TSurface;
    procedure SetItem(i : integer; sur : TSurface);
  public
    property Items[i : integer] : TSurface read GetItem write SetItem; default;
  end;

  TVertex = class(TObject)
  public
    v : TVector;
    constructor Create(x, y, z : double);
  end;

  TVertexList = class(TBaseList)
  private
    FvecCenter : TVector;
    FbRecalcCenter : boolean;
    function GetCenter: TVector;
  protected
    function GetItem(i : integer) : TVertex;
    procedure SetItem(i : integer; vx : TVertex);
  public
    constructor Create;
    property Items[i : integer] : TVertex read GetItem write SetItem; default;
    property vecCenter : TVector read GetCenter;
  end;

  TTXVertex = class(TObject)
  public
    pt : TUV;
    constructor Create(u, v : double);
  end;

  TTXVertexList = class(TBaseList)
  protected
    function GetItem(i : integer) : TTXVertex;
    procedure SetItem(i : integer; txv : TTXVertex);
  public
    property Items[i : integer] : TTXVertex read GetItem write SetItem; default;
  end;

//TODO: Untested
procedure vxlTranslate(vxl : TVertexList ; vecDelta : TVector);
procedure vxlScale(vxl : TVertexList ; sngScale : single);
function vxlDist(vxl : TVertexList ; vec : TVector) : single;
procedure vxlMove(vxl : TVertexList ; vec : TVector);

implementation

{------------------------------------------------------------------------------}
destructor TBaseList.Destroy;
begin
//  FreeAll;
  inherited;
end;

{------------------------------------------------------------------------------}
procedure TBaseList.FreeAll;
var
  i : integer;
begin
  for i := Count - 1 downto 0 do
    FreeAndNil(List[i]);
  Clear;
end;

{------------------------------------------------------------------------------}
constructor T3DMesh.Create;
begin
  inherited;
  FVertices := TVertexList.Create;
  FSurfaces := TSurfaceList.Create;
end;

{------------------------------------------------------------------------------}
destructor T3DMesh.Destroy;
begin
  inherited;
  FSurfaces.FreeAll;
  FSurfaces.Free;
  FVertices.FreeAll;
  FVertices.Free;
end;

{------------------------------------------------------------------------------}
function T3DMeshList.GetItem(i: integer): T3DMesh;
begin
  if (i < 0) or (i >= count) then
    raise EListError.CreateFmt('Sector Index is out of bounds: %d', [i]);
  Result := T3DMesh(List[i]);
end;

{------------------------------------------------------------------------------}
procedure T3DMeshList.SetItem(i: integer; msh: T3DMesh);
begin
  if (i < 0) or (i >= count) then
    raise EListError.CreateFmt('Sector Index is out of bounds: %d', [i]);
  List[i] := msh;
end;

{------------------------------------------------------------------------------}
constructor TSurface.Create;
begin
  inherited;
  FVertices := TVertexList.Create;
  FTXVertices := TTXVertexList.Create;
  iPortalSec := -1;
  iPortalSur := -1;
end;

{------------------------------------------------------------------------------}
destructor TSurface.Destroy;
begin
  inherited;
  FTXVertices.FreeAll;
  FTXVertices.Free;
  FVertices.Free;
end;

{------------------------------------------------------------------------------}
function TSurface.GetIsPortal: boolean;
begin
  Result := iPortalSec <> -1;
end;

{------------------------------------------------------------------------------}
function TSurface.GetCenter: TVector;
//var
//  i : integer;
begin
//  if not FbCenterCalculated then begin
//    for i := 0 to Vertices.Count - 1 do begin
//      FvecCenter.x := FvecCenter.x + Vertices[i].v.x;
//      FvecCenter.y := FvecCenter.y + Vertices[i].v.y;
//      FvecCenter.z := FvecCenter.z + Vertices[i].v.z;
//    end;
//    FvecCenter.x := FvecCenter.x / Vertices.Count;
//    FvecCenter.y := FvecCenter.y / Vertices.Count;
//    FvecCenter.z := FvecCenter.z / Vertices.Count;
//    FbCenterCalculated := TRUE;
//  end;

  Result := Vertices.vecCenter;
end;

{------------------------------------------------------------------------------}
function TSurface.GetNormal: TVector;
begin
  if not vecValid(FvecNormal) then
    FvecNormal := P_vec.CalcNormal(FVertices[0].v,
                                   vecCenter,
                                   FVertices[1].v);
  Result := FvecNormal;
end;

{------------------------------------------------------------------------------}
function TSurface.AddVertex(vx: TVertex ; txvx : TTXVertex): integer;
begin
  Result := FVertices.Add(vx);
  if Assigned(txvx) then
    // Create a local copy of tx vertex
    FTXVertices.Add(TTXVertex.Create(txvx.pt.u, txvx.pt.v))
  else
    FTXVertices.Add(TTXVertex.Create(0, 0));
end;

{------------------------------------------------------------------------------}
procedure TSurface.DeleteVertex(i: integer);
begin
  FVertices.Delete(i);
  FTXVertices[i].Free;
  FTXVertices.Delete(i);
end;

{------------------------------------------------------------------------------}
function TSurfaceList.GetItem(i: integer): TSurface;
begin
  if (i < 0) or (i >= count) then
    raise EListError.CreateFmt('Surface Index is out of bounds: %d', [i]);
  Result := TSurface(List[i]);
end;

{------------------------------------------------------------------------------}
procedure TSurfaceList.SetItem(i: integer; sur: TSurface);
begin
  if (i < 0) or (i >= count) then
    raise EListError.CreateFmt('Surface Index is out of bounds: %d', [i]);
  List[i] := sur;
end;

{------------------------------------------------------------------------------}
constructor TVertex.Create(x, y, z: double);
begin
  v.x := x;
  v.y := y;
  v.z := z;
end;

{------------------------------------------------------------------------------}
constructor TVertexList.Create;
begin
  FbRecalcCenter := TRUE;
end;

{------------------------------------------------------------------------------}
function TVertexList.GetItem(i: integer): TVertex;
begin
  if (i < 0) or (i >= count) then
    raise EListError.CreateFmt('Vertex Index is out of bounds: %d', [i]);
  Result := TVertex(List[i]);
end;

{------------------------------------------------------------------------------}
procedure TVertexList.SetItem(i: integer; vx: TVertex);
begin
  if (i < 0) or (i >= count) then
    raise EListError.CreateFmt('Vertex Index is out of bounds: %d', [i]);
  List[i] := vx;
  FbRecalcCenter := TRUE;
end;

{------------------------------------------------------------------------------}
function TVertexList.GetCenter: TVector;
var
  i : integer;
begin
//  if FbRecalcCenter then begin
    FvecCenter := P_vec.vec(0, 0, 0);
    for i := 0 to Count - 1 do begin
      FvecCenter.x := FvecCenter.x + Items[i].v.x;
      FvecCenter.y := FvecCenter.y + Items[i].v.y;
      FvecCenter.z := FvecCenter.z + Items[i].v.z;
    end;
    FvecCenter.x := FvecCenter.x / Count;
    FvecCenter.y := FvecCenter.y / Count;
    FvecCenter.z := FvecCenter.z / Count;
    FbRecalcCenter := FALSE;
//  end;
  Result := FvecCenter;
end;

{------------------------------------------------------------------------------}
constructor TTXVertex.Create(u, v: double);
begin
  pt.u := u;
  pt.v := v;
end;

{------------------------------------------------------------------------------}
function TTXVertexList.GetItem(i: integer): TTXVertex;
begin
  if (i < 0) or (i >= count) then
    raise EListError.CreateFmt('Texture Vertex Index is out of bounds: %d', [i]);
  Result := TTXVertex(List[i]);
end;

{------------------------------------------------------------------------------}
procedure TTXVertexList.SetItem(i: integer; txv: TTXVertex);
begin
  if (i < 0) or (i >= count) then
    raise EListError.CreateFmt('Texture Vertex Index is out of bounds: %d', [i]);
  List[i] := txv;
end;

{------------------------------------------------------------------------------}
procedure vxlTranslate(vxl : TVertexList ; vecDelta : TVector);
var
  i : integer;
begin
  for i := 0 to vxl.Count - 1 do
    vxl[i].v := vecAdd(vxl[i].v, vecDelta);
end;

{------------------------------------------------------------------------------}
procedure vxlScale(vxl : TVertexList ; sngScale : single);
var
  vecOldPos : TVector;
  i : integer;
begin
  vecOldPos := vxl.vecCenter;
  vxlMove(vxl, P_vec.vec(0, 0, 0));
  for i := 0 to vxl.Count - 1 do
    vxl[i].v := vecMultiply(vxl[i].v, vec(sngScale, sngScale, sngScale));
  vxlMove(vxl, vecOldPos);
end;

{------------------------------------------------------------------------------}
function vxlDist(vxl : TVertexList ; vec : TVector) : single;
var
  vecD : Tvector;
begin
  vecD := vecSubtract(vxl.vecCenter, vec);
  Result := sqrt(vecD.x*vecD.x + vecD.y*vecD.y + vecD.z*vecD.z);
end;

{------------------------------------------------------------------------------}
procedure vxlMove(vxl : TVertexList ; vec : TVector);
var
  vecD : Tvector;
begin
  vecD := vecSubtract(vxl.vecCenter, vec);
  vxlTranslate(vxl, P_vec.vec(-vecD.x, -vecD.y, -vecD.z));
  asm nop end;
end;

end.
