(*
 * Copyright (c) 2002 Andr di Geronimo (adg@home.se)
 *)
 
unit O_jkl;

interface

uses
  Windows, Classes, SysUtils,
  P_str, O_geo, O_ComEnt, P_vec;

type
  TSurfaceFlag = (Floor, CogLinked, Impassable, AiCannotWalkOnFloor, DoubleTex,
                  HalfTex, QTex, NoFallingDamage, Unk1, Sky, Sky2, Scrolling, Icy, Ice2,
                  Magsealed, Metal, Deep, Shallow, Dirt, Deep2);
  TSurfaceFlags = set of TSurfaceFlag;

  TCallbackStatus = procedure(iStart, iEnd : integer);

  TSector = T3DMesh;

  TJKLevel = class(TObject)
  private
    FsName : string;
    FstlWorldMaterials : TStringList;
    FstlUsedMaterials : TStringList;
    FlstWorldVertices : TVertexList;
    FlstWorldTXVertices : TTXVertexList;
    FlstWorldPortals : TList;
    FlstWorldSurfaces : TSurfaceList;
    FlstWorldSectors : T3DMeshList;
    FlstWorldThings : TStringList;
    FlstWorldLights : TList;
    procedure SetName(const Value: string);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function LoadFromFile(const sFilename : string): boolean;
    function LoadFromStl(stl : TStringList): boolean;
    function LoadJKLFromFile(const sFilename : string ;
      procStatus : TCallbackStatus = nil): boolean;
    function LoadJKLFromStl(stl : TStringList ;
      procStatus : TCallbackStatus = nil): boolean;
    property Name : string read FsName write SetName;
    property lstWorldSectors : T3DMeshList read FlstWorldSectors;
    property lstWorldThings : TStringList read FlstWorldThings;
    property lstWorldLights : TList read FlstWorldLights;
    property lstUsedMaterials : TStringList read FstlUsedMaterials;
  end;

implementation

{------------------------------------------------------------------------------}
constructor TJKLevel.Create;
begin
  FstlWorldMaterials := TStringList.Create;
  FstlUsedMaterials := TStringList.Create;
  FlstWorldVertices := TVertexList.Create;
  FlstWorldTXVertices := TTXVertexList.Create;
  FlstWorldPortals := TList.Create;
  FlstWorldSurfaces := TSurfaceList.Create;
  FlstWorldSectors  := T3DMeshList.Create;
  FlstWorldThings := TStringList.Create;
  FlstWorldLights := TList.Create;
end;

{------------------------------------------------------------------------------}
destructor TJKLevel.Destroy;
begin
  Clear;
  FstlWorldMaterials.Free;
  FstlUsedMaterials.Free;
  FlstWorldPortals.Free;
  FlstWorldSurfaces.Free;
  FlstWorldVertices.Free;
  FlstWorldTXVertices.Free;
  FlstWorldSectors.Free;
  FlstWorldThings.Free;
  inherited;
end;

{------------------------------------------------------------------------------}
procedure TJKLevel.SetName(const Value: string);
begin
  FsName := ChangeFileExt(ExtractFilename(Value), '')
end;

{------------------------------------------------------------------------------}
procedure TJKLevel.Clear;
var
  i : integer;
  pThg : PThing;
  pLit : PLight;
begin
  FstlWorldMaterials.Clear;
  FstlUsedMaterials.Clear;
  FlstWorldVertices.FreeAll;
  FlstWorldVertices.Clear;
  FlstWorldTXVertices.FreeAll;
  FlstWorldTXVertices.Clear;
  FlstWorldSectors.FreeAll;
  FlstWorldSectors.Clear;

  for i := 0 to FlstWorldThings.Count - 1 do begin
    pThg := PThing(FlstWorldThings.Objects[i]);
    Dispose(pThg);
  end;
  FlstWorldThings.Clear;

  for i := 0 to FlstWorldLights.Count - 1 do begin
    pLit := PLight(FlstWorldLights[i]);
    Dispose(pLit);
  end;
  FlstWorldLights.Clear;
end;

{------------------------------------------------------------------------------}
function TJKLevel.LoadFromFile(const sFilename: string): boolean;
var
  stl : TStringList;
begin
  stl := TStringList.Create;
  try
    stl.LoadFromFile(sFilename);
    Result := LoadFromStl(stl);
  finally
    stl.Free;
  end;
end;

{------------------------------------------------------------------------------}
function TJKLevel.LoadFromStl(stl: TStringList): boolean;
var
  iCurLine : integer;
  sLine : string;

  {--------------------------------------------------}
  procedure ReadLn;
  begin
    sLine := stl[iCurLine];
    if Pos('#', sLine) > 0 then
      sLine := Copy(sLine, Pos('#', sLine), Length(sLine));
    sLine := TrimLeft(sLine);
    inc(iCurLine);
  end;

  {--------------------------------------------------}
  function GetValue(const sKey : string) : string;
  var
    s : string;
    b : boolean;
  begin
    repeat
      ReadLn;
      s := sLine;
      b := Pos(sKey, s) > 0;
      if b then begin
        Delete(s, 1, Pos(' ', s));
        Result := s;
      end;
    until b or (iCurLine >= stl.Count-1);
  end;

var
  iSecCount : integer;
  iSec : integer;
  sec : TSector;
  iVxCount : integer;
  iVx : integer;
  iVxIndex : integer;
  vx : TVertex;
  iSurCount : integer;
  iSur : integer;
  sur : TSurface;
begin
  Clear;
  iCurLine := 0;
  with stl do begin
    // Sectors
    iSecCount := StrToInt(GetValue('SECTORS'));
    for iSec := 0 to iSecCount - 1 do begin

      // Sector
      GetValue('SECTOR ' + IntToStr(iSec));
      with FlstWorldSectors do
        sec := Items[Add(TSector.Create)];

      // Vertices
      iVxCount := StrToInt(GetValue('VERTICES'));
      for iVx := 0 to iVxCount - 1 do begin
        GetValue(IntToStr(iVx) + ':');
        with sec.Vertices do
          vx := Items[Add(TVertex.Create(0, 0, 0))];
        vx.v.x := StrToFloat(strTokenAt(sLine, ' ', 1));
        vx.v.y := StrToFloat(strTokenAt(sLine, ' ', 2));
        vx.v.z := StrToFloat(strTokenAt(sLine, ' ', 3));
      end;

      // Surfaces
      iSurCount := StrToInt(GetValue('SURFACES'));
      for iSur := 0 to iSurCount - 1 do begin
        GetValue('SURFACE ' + IntToStr(iSur));
        with sec.Surfaces do
          sur := Items[Add(TSurface.Create)];
        GetValue('PORTAL');
        sur.iPortalSec := StrToInt(strTokenAt(sLine, ' ', 1));
        sur.iPortalSur := StrToInt(strTokenAt(sLine, ' ', 2));

        // Surface vertices
        iVxCount := StrToInt(GetValue('SURFACE_VERTICES'));
        for iVx := 0 to iVxCount - 1 do begin
          GetValue(IntToStr(iVx) + ':');
          iVxIndex := StrToInt(strTokenAt(sLine, ' ', 1));
          sur.AddVertex(sec.Vertices[iVxIndex]);
          sur.TXVertices[iVx].pt.u := StrToFloat(strTokenAt(sLine, ' ', 2));
          sur.TXVertices[iVx].pt.v := StrToFloat(strTokenAt(sLine, ' ', 3));
        end;
      end;

    end;
  end;
  Result := TRUE;
end;

{------------------------------------------------------------------------------}
function TJKLevel.LoadJKLFromFile(const sFilename: string ;
  procStatus : TCallbackStatus): boolean;
var
  stl : TStringList;
begin
  stl := TStringList.Create;
  with stl do
  try
    stl.LoadFromFile(sFilename);
    Result := LoadJKLFromStl(stl, procStatus);
    Name := sFilename;
  finally
    stl.Free;
  end;
end;

{------------------------------------------------------------------------------}
function TJKLevel.LoadJKLFromStl(stl: TStringList ;
  procStatus : TCallbackStatus): boolean;
var
  iCurLine : integer;
  sCurLine : string;
  sNextLine : string;

  {--------------------------------------------------}
  procedure ReadLn;
  begin
    sCurLine := stl[iCurLine];
    if iCurLine+1 < stl.Count then
      sNextLine := stl[iCurLine+1];
    // Strip comments
    if Pos('#', sCurLine) > 0 then
      Delete(sCurLine, Pos('#', sCurLine), Length(sCurLine));
    // Tab => Space
    if Pos(#9, sCurLine) > 0 then
      sCurLIne := strReplaceChar(sCurLIne, #9, ' ');
    sCurLine := strSingleSpace(sCurLine);
    sCurLine := Trim(sCurLine);
    if Assigned(procStatus) then
      procStatus(iCurLine, stl.Count);
    inc(iCurLine);
  end;

  {--------------------------------------------------}
  function SkipToSection(const sSecToken : string ;
    bRaiseNotFound : boolean = TRUE) : boolean;
  var
    bFoundSection : boolean;
  begin
    repeat
      bFoundSection :=
        (Pos('SECTION:', UpperCase(sCurLine)) > 0) and
        (Pos(UpperCase(sSecToken), UpperCase(sCurLine)) > 0);
      if not bFoundSection then
        ReadLn;
    until bFoundSection or (iCurLine >= stl.Count - 1);
    Result := bFoundSection;
    if not Result and bRaiseNotFound then
      raise Exception.Create('Error reading JKL file, ' + sSecToken + ' section not found.');
  end;

  {--------------------------------------------------}
  function SkipToToken(const sToken : string) : boolean;
  var
    bFoundToken : boolean;
  begin
    repeat
      bFoundToken :=
        Pos(UpperCase(sToken), UpperCase(sCurLine)) = 1;
      if not bFoundToken then
        ReadLn;
    until bFoundToken or (iCurLine >= stl.Count - 1);
    Result := bFoundToken;
    if not Result then
      raise Exception.Create('Error reading JKL file, expected token ''' + sToken + ''' not found.');
  end;

var
  i, j : integer;
  iCount, iCount2 : integer;
  iVx, iTxvx : integer;
  sVxPair : string;
  sur : TSurface;
  sec : TSector;
  vecNorm : TVector;
  pThg : PThing;
  pLit : PLight;
begin
  Clear;
  iCurLine := 0;

  SkipToSection('MATERIALS');
  
  SkipToToken('WORLD MATERIALS');
  iCount := StrToInt(strAlpha(sCurLine));
  for i := 0 to iCount - 1 do begin
    SkipToToken(IntToStr(i)+':');
    FstlWorldMaterials.Add(strTokenAt(sCurLine, ' ', 1));
    // In some LEC levels the iCount reported does not seem to be correct,
    // the token "end" seems to indicate the _real_ end of the list.
    if LowerCase(sNextLine) = 'end' then
      Break;
  end;

  SkipToSection('GEORESOURCE');

  SkipToToken('WORLD VERTICES');
  iCount := StrToInt(strAlpha(sCurLine));
  for i := 0 to iCount - 1 do begin
    SkipToToken(IntToStr(i)+':');
    FlstWorldVertices.Add(
      TVertex.Create(StrToFloat(strTokenAt(sCurLine, ' ', 1)),
                     StrToFloat(strTokenAt(sCurLine, ' ', 2)),
                     StrToFloat(strTokenAt(sCurLine, ' ', 3))));
  end;

  SkipToToken('WORLD TEXTURE VERTICES');
  iCount := StrToInt(strAlpha(sCurLine));
  for i := 0 to iCount - 1 do begin
    SkipToToken(IntToStr(i)+':');
    FlstWorldTXVertices.Add(
      TTXVertex.Create(StrToFloat(strTokenAt(sCurLine, ' ', 1)),
                       StrToFloat(strTokenAt(sCurLine, ' ', 2))));
  end;

  SkipToToken('WORLD ADJOINS');
  iCount := StrToInt(strAlpha(sCurLine));
  for i := 0 to iCount - 1 do begin
    SkipToToken(IntToStr(i)+':');
    FlstWorldPortals.Add(Pointer(StrToInt(strTokenAt(sCurLine, ' ', 2))));
  end;

  SkipToToken('WORLD SURFACES');
  iCount := StrToInt(strAlpha(sCurLine));
  for i := 0 to iCount - 1 do begin
    SkipToToken(IntToStr(i)+':');
    sur := TSurface.Create;
    FlstWorldSurfaces.Add(sur);

    j := StrToInt(strTokenAt(sCurLine, ' ', 1)); 
    if j <> -1 then // Material not required
      sur.sMaterial := FstlWorldMaterials[j]
    else
      sur.sMaterial := 'dflt.mat';

    if FstlUsedMaterials.IndexOf(LowerCase(sur.sMaterial)) = -1 then
      FstlUsedMaterials.Add(sur.sMaterial);

    sur.Flags := StrToInt(strTokenAt(sCurLine, ' ', 2));
    sur.iAdjoin := StrToInt(strTokenAt(sCurLine, ' ', 7));

    for j := 1 to StrToInt(strTokenAt(sCurLine, ' ', 9)) do begin
      sVxPair := strTokenAt(sCurLine, ' ', 9 + j);
      // sVxPair ~ "Vx,TxVx"
      iVx := StrToInt(Copy(sVxPair, 1, Pos(',', sVxPair)-1));
      iTxVx := StrToInt(Copy(sVxPair, Pos(',', sVxPair)+1, Length(sVxPair)));
      if iTxVx <> -1 then
        sur.AddVertex(FlstWorldVertices[iVx], FlstWorldTXVertices[iTxVx])
      else
        sur.AddVertex(FlstWorldVertices[iVx]);
    end;
  end;

  // Surface Normals
  // Same iCount as world surfaces
  for i := 0 to iCount - 1 do begin
    SkipToToken(IntToStr(i)+':');
    sur := FlstWorldSurfaces[i];
    vecNorm.x := StrToFloat(strTokenAt(sCurLine, ' ', 1));
    vecNorm.y := StrToFloat(strTokenAt(sCurLine, ' ', 2));
    vecNorm.z := StrToFloat(strTokenAt(sCurLine, ' ', 3));
    sur.vecNormal := vecNorm;
  end;

  SkipToSection('SECTORS');

  SkipToToken('WORLD SECTORS');
  iCount := StrToInt(strAlpha(sCurLine));
  for i := 0 to iCount - 1 do begin
    SkipToToken('SECTOR ' + IntToStr(i));
    sec := TSector.Create;
    FlstWorldSectors.Add(sec);

    //TODO:Read vertices section? SkipToToken('VERTICES: ' + IntToStr(i));

    SkipToToken('SURFACES');
    j := StrToInt(strTokenAt(sCurLine, ' ', 1));
    iCount2 := StrToInt(strTokenAt(sCurLine, ' ', 2));
    for j := j to (j + iCount2)-1 do
      sec.Surfaces.Add(FlstWorldSurfaces[j]);
  end;

  SkipToSection('THINGS');            
  SkipToToken('WORLD THINGS');
  iCount := StrToInt(strAlpha(sCurLine));
  for i := 0 to iCount - 1 do begin
    SkipToToken(IntToStr(i) + ':');
    New(pThg);
    pThg^.sName := strTokenAt(sCurLine, ' ', 1);
    pThg^.Pos.x := StrToFloat(strTokenAt(sCurLine, ' ', 3));
    pThg^.Pos.y := StrToFloat(strTokenAt(sCurLine, ' ', 4));
    pThg^.Pos.z := StrToFloat(strTokenAt(sCurLine, ' ', 5));
    pThg^.Rot.x := StrToFloat(strTokenAt(sCurLine, ' ', 6));
    pThg^.Rot.y := StrToFloat(strTokenAt(sCurLine, ' ', 7));
    pThg^.Rot.z := StrToFloat(strTokenAt(sCurLine, ' ', 8));
    FlstWorldThings.AddObject(pThg^.sName, TObject(pThg));

    if LowerCase(sNextLine) = 'end' then
      Break;
  end;

  // Not required; only written by editors (JED, etc)
  if SkipToSection('LIGHTS', FALSE) then begin
    SkipToToken('EDITOR LIGHTS');
    iCount := StrToInt(strAlpha(sCurLine));
    for i := 0 to iCount - 1 do begin
      SkipToToken(IntToStr(i) + ':');
      New(pLit);
      pLit^.iType := StrToInt(strTokenAt(sCurLine, ' ', 1));
      pLit^.Pos.x := StrToFloat(strTokenAt(sCurLine, ' ', 2));
      pLit^.Pos.y := StrToFloat(strTokenAt(sCurLine, ' ', 3));
      pLit^.Pos.z := StrToFloat(strTokenAt(sCurLine, ' ', 4));
      pLit^.fRange := StrToFloat(strTokenAt(sCurLine, ' ', 5));
      pLit^.fIntensity := StrToFloat(strTokenAt(sCurLine, ' ', 6));

      FlstWorldLights.Add(TObject(pLit));

      if LowerCase(sNextLine) = 'end' then
        Break;
    end;
  end;
  
  Result := TRUE;
end;

end.
