(*
 * Copyright (c) 2002 Andr di Geronimo (adg@home.se)
 *)
 
unit O_jkt;

interface

uses
  Windows, Classes, SysUtils, Graphics;

type
  TCMPHeader = packed record
    sig : array[0..3] of char; {'CMP '}
    twenty : longint;
    HasTransparency : Longint;
    stuff:array[1..52] of byte;
  end;

  TCMPPalette = array[0..255] of
    {packed }record
      r,
      g,
      b : byte;
    end;

{$I dflt.cmp.inc}

type
  PLongWordArray = ^TLongWordArray;
  TLongWordArray = array[0..16383] of LongWord;

  TMatHeader = packed record
    tag:array[0..3] of char;     {'MAT ' - notice space after MAT}
    ver:Longint;                 {Apparently - version = 0x32 ('2')}
    mat_Type:Longint;            {0 = colors(TColorHeader) , 1= ?, 2= texture(TTextureHeader)}
    NumOfTextures:Longint;       {number of textures or colors}
    NumOfTextures1: Longint;     { In color MATs, it's 0, in TX ones, it's equal to numOfTextures }
    pad1:Longint;                { = 0 }
    pad2:LongInt;                { = 8 }
    pads:array[0..11]of longint; {unknown. Some pad?}
  end;

  TColorHeader = packed record
    textype:longint;            {0 = color, 8= texture}
    colornum:longint;           {Color index from the CMP palette}
    pads:array[0..3]of Longint; {each = 0x3F800000 (check cmp header )}
  end;

  TTextureHeader = packed record
    textype:longint;             {0 = color, 8= texture}
    colornum:longint;            {unknown use}
    pads:array[0..3]of Longint;  {each longint = 0x3F800000 (check cmp header )}
    pads1:array[0..1]of Longint; {unknown}
    pad2:Longint;                {=0xBFF78482}
    CurrentTXNum:Longint         {number of corresponding texture, beginning with 0, ranging to NumOfTextures-1}
  end;

  TTextureData = packed record
    SizeX:Longint;              {horizontal size of first MipMap, must be divisable by 2}
    SizeY:Longint;              {Vertical size of first MipMap ,must be divisable by 2}
    Pad:array[0..2]of LongInt;  {padding = 0 }
    NumMipMaps:LongInt;         {Number of mipmaps in texture largest one first.}
  end;
  

  TTGAHeader = packed record
    ID : byte;           {0, 1..255}
    ColorMapType : byte; {0, 1..255}
    ImageType : byte;    { 0 No Image Data Included
                           1 Uncompressed, Color-mapped Image
                           2 Uncompressed, True-color Image
                           3 Uncompressed, Black-and-white Image
                           9 Run-length encoded, Color-mapped Image
                          10 Run-length encoded, True-color Image
                          11 Run-length encoded, Black-and-white Image}
  end;

  TTGAColorMapHeader = packed record
    FirstEntryIndex : smallint;
    Length : smallint;
    EntrySize : byte;
  end;

  TTGAImageDataHeader = packed record
    XOrigin : smallint;
    YOrigin : smallint;
    Width : smallint;
    Height : smallint;
    PixelDepth : byte;      {8, 16, 24, 32 etc}
    ImageDescriptor : byte; { Bits 3-0: These bits specify the number of attribute bits per
                              pixel. In the case of the TrueVista, these bits indicate
                              the number of bits per pixel which are designated as
                              Alpha Channel bits. For the ICB and TARGA
                              products, these bits indicate the number of overlay
                              bits available per pixel. See field 24 (Attributes
                              Type) for more information. }
                            { Bits 5 & 4:These bits are used to indicate the order in which
                              pixel data is transferred from the file to the screen.
                              Bit 4 is for left-to-right ordering and bit 5 is for top-to-
                              bottom ordering as shown below.
                              Table 2 - Image Origin
                              Bits 7 & 6:Must be zero to insure future compatibility.


                              Screen destination    Image Origin
                              of first pixel        bit 5 bit 4
                              ---------------------------------
                              bottom left           0     0
                              bottom right          0     1
                              top left              1     0
                              top right             1     1 }
                             { Bits 7 & 6: 0 }
  end;

  TTGAFooter = packed record
    ExtensionAreaOffset : longint;
    DeveloperDirectoryOffset : longint;
    Signature : array[0..15] of Char;   {= 'TRUEVISION-XFILE'}
    Reserved : Char;                    {= '.'}
    Eof : byte;                         {= 0}
  end;


type
  TJKTexture = class(TObject)
  private
    FaPalette: TCMPPalette;
    FiWidth : integer;
    FiHeight : integer;
    FBitmap : TBitmap;
    procedure SetWidth(const Value: integer);
    procedure SetHeight(const Value: integer);
  public
    constructor Create;
    destructor Destroy; override;
    function LoadFromStream(stm : TStream) : boolean;
    function LoadFromFile(const sFilename : string) : boolean;
    function SaveAsJPEGToFile(const sFilename : string) : boolean;
    function SaveAsTGAToFile(const sFilename : string) : boolean;
    function SaveToFile(const sFilename : string) : boolean;
    property Palette : TCMPPalette read FaPalette write FaPalette;
    property Bitmap : TBitmap read FBitmap;
    property Width : integer read FiWidth write SetWidth;
    property Height : integer read FiHeight write SetHeight;
  end;

  TColorMap = class(TObject)
  private
    FaPalette : TCMPPalette;
  public
    function LoadPaletteFromStream(stm : TStream) : boolean;
    function LoadPaletteFromFile(const sFilename : string) : boolean;
    property Palette : TCMPPalette read FaPalette write FaPalette;
  end;

function MatToJPG(const sInput, sOutput : string ; aPalette : TCMPPalette) : boolean;
function MatToTGA(const sInput, sOutput : string ; aPalette : TCMPPalette) : boolean;

implementation

uses
  JPEG;

{------------------------------------------------------------------------------}
constructor TJKTexture.Create;
begin
  FBitmap := TBitmap.Create;
  FBitmap.PixelFormat := pf32bit;
  FaPalette := DEFAULT_PALETTE;
end;

{------------------------------------------------------------------------------}
destructor TJKTexture.Destroy;
begin
  FBitmap.Free;
  inherited;
end;

{------------------------------------------------------------------------------}
procedure TJKTexture.SetWidth(const Value: integer);
begin
  FiWidth := Value;
  FBitmap.Width := FiWidth;
end;

{------------------------------------------------------------------------------}
procedure TJKTexture.SetHeight(const Value: integer);
begin
  FiHeight := Value;
  FBitmap.Height := FiHeight;
end;

{------------------------------------------------------------------------------}
function TJKTexture.LoadFromStream(stm: TStream): boolean;
var
  recHead : TMatHeader;
  recTexHead : TTextureHeader;
  recTexData : TTextureData;
  x, y : integer;
  pScanLine : PLongWordArray;
  pBuff : PByteArray;
begin
  Result := TRUE;
  stm.Position := 0;
  stm.ReadBuffer(recHead, SizeOf(recHead));
  if recHead.mat_Type = 2 then begin
    stm.ReadBuffer(recTexHead, SizeOf(recTexHead));
    // Skip the following Texture Headers
    stm.Seek((recHead.NumOfTextures-1) * SizeOf(TTextureHeader), soFromCurrent);

    if recTexHead.textype = 8 then begin // 8bit
      stm.ReadBuffer(recTexData, SizeOf(recTexData));
      // Skip the following Texture Headers
      stm.Seek((recHead.NumOfTextures-1) * SizeOf(TTextureData), soFromCurrent);

      Width := recTexData.SizeX;
      Height := recTexData.SizeY;

      // pixel data
      pBuff := AllocMem(Width);
      try
        for y := 0 to Height - 1 do begin
          pScanLine := FBitmap.ScanLine[y];
          stm.ReadBuffer(pBuff^, Width);
          for x := 0 to Width - 1 do begin
            pScanLine^[x] :=
              RGB(Palette[pBuff^[x]].b,  // swap R and B
                  Palette[pBuff^[x]].g,
                  Palette[pBuff^[x]].r);
          end;
        end;
      finally
        FreeMem(pBuff);
      end;
    end;
    
  end;
end;

{------------------------------------------------------------------------------}
function TJKTexture.LoadFromFile(const sFilename: string): boolean;
var
  stm : TFileStream;
begin
  if FileExists(sFilename) then begin
    stm := TFileStream.Create(sFilename, fmOpenRead);
    try
      Result := Self.LoadFromStream(stm);
    finally
      stm.Free;
    end;
  end
  else
    Result := FALSE;
end;

{------------------------------------------------------------------------------}
function TJKTexture.SaveAsJPEGToFile(const sFilename: string): boolean;
var
  jpg : TJPEGImage;
begin
  jpg := TJPEGImage.Create;
  try
    jpg.Assign(Bitmap);
    jpg.SaveToFile(sFilename);
  finally
    jpg.Free;
  end;
  Result := TRUE;
end;

{------------------------------------------------------------------------------}
function TJKTexture.SaveAsTGAToFile(const sFilename: string): boolean;
var
  recHead : TTGAHeader;
  recColorMapHead : TTGAColorMapHeader;
  recImageHead : TTGAImageDataHeader;
  recFooter : TTGAFooter;
  stm : TFileStream;
  y : integer;
  pPixelsRow : PLongWordArray;
begin
  stm := TFileStream.Create(sFilename, fmCreate);
  try
    stm.Position := 0;

    // TGA header
    with recHead do begin
      ID := 0;
      ColorMapType := 0;
      ImageType := 2;
    end;
    stm.WriteBuffer(recHead, SizeOf(recHead));

    // ColorMap header
    // Not supported, fill with 0s
    FillChar(recColorMapHead, SizeOf(recColorMapHead), 0);
    stm.WriteBuffer(recColorMapHead, SizeOf(recColorMapHead));

    // Image header
    with recImageHead do begin
      XOrigin := 0;
      YOrigin := 0;
      Width := FBitmap.Width;
      Height := FBitmap.Height;
      PixelDepth := 32;
      ImageDescriptor := 0 or (1 shl 5); // set bit 5 (= Start at Top Left)
    end;
    stm.WriteBuffer(recImageHead, SizeOf(recImageHead));

    // Image data, 32-bit
    for y := 0 to FBitmap.Height - 1 do begin
      pPixelsRow := FBitmap.ScanLine[y];
      stm.WriteBuffer(pPixelsRow^, FBitmap.Width * 4);
    end;

    // TGA footer
    with recFooter do begin
      ExtensionAreaOffset := 0;
      DeveloperDirectoryOffset := 0;
      Signature := 'TRUEVISION-XFILE';
      Reserved := '.';
      Eof := 0;
    end;
    stm.WriteBuffer(recFooter, SizeOf(recFooter));

    Result := TRUE;
  finally
    stm.Free;
  end;
end;

{------------------------------------------------------------------------------}
function TJKTexture.SaveToFile(const sFilename: string): boolean;
var
  sExt : string;
begin
  sExt := LowerCase(ExtractFileExt(sFilename));
  if (sExt = '.jpg') or (sExt = '.jpeg') then
    Result := SaveAsJPEGToFile(sFilename)
  else if sExt = '.tga' then
    Result := SaveAsTGAToFile(sFilename)
  else
    raise Exception.Create('Unsupported file extension: ''' + sExt + '''');
end;

{------------------------------------------------------------------------------}
function MatToJPG(const sInput, sOutput : string ; aPalette : TCMPPalette) : boolean;
begin
  Result := TRUE;
  with TJKTexture.Create do
  try
    Palette := aPalette;
    LoadFromFile(sInput);
    SaveAsJPEGToFile(sOutput);
  except
    Free;
    Result := FALSE;
  end;
end;

{------------------------------------------------------------------------------}
function MatToTGA(const sInput, sOutput : string ; aPalette : TCMPPalette) : boolean;
begin
  Result := TRUE;
  with TJKTexture.Create do
  try
    Palette := aPalette;
    LoadFromFile(sInput);
    SaveAsTGAToFile(sOutput);
  except
    Free;
    Result := FALSE;
  end;
end;

{------------------------------------------------------------------------------}
function TColorMap.LoadPaletteFromStream(stm: TStream): boolean;
var
  recHead : TCMPHeader;
//  pal : TCMPPalette;
begin
  stm.Position := 0;
  stm.ReadBuffer(recHead, SizeOf(recHead));
  stm.ReadBuffer(FaPalette, SizeOf(FaPalette));
  // Not intrested in the light tables or the transparency tables
  Result := TRUE;
end;

{------------------------------------------------------------------------------}
function TColorMap.LoadPaletteFromFile(const sFilename: string): boolean;
var
  stm : TFileStream;
begin
  if FileExists(sFilename) then begin
    stm := TFileStream.Create(sFilename, fmOpenRead);
    try
      Result := LoadPaletteFromStream(stm);
    finally
      stm.Free;
    end;
  end
  else
    Result := FALSE;
end;

end.
